import logging
import base64
import os
import requests
from datetime import datetime
from urllib.parse import urlparse, unquote
from . import BaseTool
from .. import attachments
logger = logging.getLogger('sigmund')


class download(BaseTool):
    """Download a file or webpage from the internet. The downloaded file or webpage will be saved as attachment, and can be read later.·"""
    
    arguments = {
        "url": {
            "type": "string",
            "description": "The url of a file of webpage",
        }
    }
    required_arguments = ["url"]    

    def _download(self, url):
        """Download URL and return a filename, content tuple."""
        try:
            response = requests.get(url, stream=True)
            response.raise_for_status()
            # Attempt to get the filename from the Content-Disposition header
            content_disposition = response.headers.get('Content-Disposition')
            if content_disposition:
                # If there's a filename in the header, extract it
                filename = re.findall('filename="?(.+)"?', 
                                      content_disposition)[0]
            else:
                # Parse the URL to get a base filename
                parsed_url = urlparse(url)
                if parsed_url.path:
                    # Unquote URL to decode %20, etc. to actual characters
                    filename = unquote(os.path.basename(parsed_url.path))
                else:
                    # Use the domain name and a timestamp as the filename
                    filename = parsed_url.netloc.replace('.', '_') + '_' \
                        + datetime.now().strftime('%Y%m%d%H%M%S')
            # Check if the filename is empty or just a slash, use default name
            if not filename or filename == '/':
                filename = 'download_' \
                    + datetime.now().strftime('%Y%m%d%H%M%S')
            # Add an appropriate file extension if missing
            if not os.path.splitext(filename)[1]:
                # Default to .html for web pages without a clear file type
                filename += '.html'  
            content = response.content
            return filename, content
        except Exception as e:
            logger.error(f"Failed to download the file from {url}: {e}")
            raise
            
    def __call__(self, url):
        filename, content = self._download(url)
        description = attachments.describe_file(filename, content,
                                                self._sigmund.condense_model)
        attachment_data = {
            'filename': filename,
            'content': base64.b64encode(content).decode('utf-8'),
            'description': description
        }
        self._sigmund.database.add_attachment(attachment_data)
        return f'''I have downloaded {filename} and added it to my attachments.''', \
            '', False


import logging
# import re
import json
import functools
from typing import Optional, Tuple
logger = logging.getLogger('sigmund')


class BaseTool:
    """A base class for tools that process an AI reply."""
    
    def __init__(self, sigmund):
        self._sigmund = sigmund
        
    @property
    def tool_spec(self):
        """The tool spec should corresond to the OpenAI specification for
        function tools.
        """
        return {
            "name": self.__class__.__name__,
            "description": self.__doc__,
            "parameters": {
                "type": "object",
                "properties": self.arguments,
                "required": self.required_arguments,
            }
        }
        
    @property
    def name(self):
        return self.__class__.__name__
    
    def bind(self, args: str) -> callable:
        """Returns a callable that corresponds to a tool function called with
        a string of arguments, which should be in JSON format. The callable 
        itself returns a (message: str, result: dict, needs_feedback: bool)
        tuple, where message is an informative text message as generated by
        the tool, result is a dict with name, args, and content keys that
        correspond to the name and arguments of the function and the result of
        the tool call. needs_feedback indicates whether the model should be
        called again to provide feedback based on the tool result.
        """
        def bound_tool_function():
            try:
                message, result, needs_feedback = self(**json.loads(args))
            except Exception as e:
                message = 'Failed to run tool'
                result = f'The following error occurred while trying to run tool:\n\n{e}'
                needs_feedback = True
            result = {'name': self.name,
                      'args': args,
                      'content': result}
            return message, result, needs_feedback
        return bound_tool_function
        
    def __call__(self) -> Tuple[str, Optional[str], bool]:
        """Should be implemented in a tool with additional arguments that
        match tool specification.
        """
        raise NotImplementedError()


import os
import re
from . import utils
import logging
logger = logging.getLogger('sigmund')
page_title = 'Sigmund AI'
ai_name = 'Sigmund'

# SERVER
#
# The external server address, that is, the URL that users visit
server_url = os.environ.get('FLASK_SERVER_URL', 'http://127.0.0.1:5000')
# The port at which the flask server is running internally. This can be
# different from the server URL if the app is running behind a proxy that
# redirects
flask_port = int(os.environ.get('FLASK_PORT', 5000))
# The flask host arhument where all 0s means listen to all incoming addresses
flask_host = os.environ.get('FLASK_HOST', '0.0.0.0')
# The secret key is used for logging in. This should be a long and arbitrary
# string that is hard to guess. This should not be shared
flask_secret_key = os.environ.get('FLASK_SECRET_KEY', '0123456789ABCDEF')
# A secret salt that is used to encrypt the messages on disk in combination
# with the user password
encryption_salt = os.environ.get('SIGMUND_ENCRYPTION_SALT', '0123456789ABCDEF').encode()

# FILES AND FOLDERS
#
# Contains the encrypted message history
sessions_folder = 'sessions'
# Contains the encrypted uploads
uploads_folder = 'uploads'

# PROMPT HISTORY
#
# The maximum length of a prompt in characters. If the prompt exceeds this 
# length, the start will be summarized.
max_prompt_length = 20000
# The length of the prompt to be summarized.
condense_chunk_length = 10000

# ATTACHMENTS
#
# The maximum length of the text representation of an attachment that is used
# to generate a description
max_text_representation_length = 100000
# The maximum length of the text representation that is not summarized. Beyond
# this, the text is summarized rather than stored in its entirety.
text_respresentation_summary_threshold = 30000

# LIBRARY INDEXING
#
# The number of documents that are indexed at once and the delay after each
# chunk. This is to avoid exceeded the rate limits of the OpenAI API
chunk_size = 100
chunk_throttle = .1

# MESSAGES
# The maximum length of a user message
max_message_length = 10000
# A fixed welcome message
welcome_message_with_search = '''Nice to meet you! I am Sigmund, your friendly OpenSesame assistant! How can I help you?

<div class="message-notification">I am currently functioning as an OpenSesame expert. <a href="/user_question_guide">Learn how to ask effective questions.</a> If you want to discuss things that are not related to OpenSesame, <a href="/api/conversation/new/without_search">click here</a> to turn me into a regular chatbot that has access to a wider range of tools.</div>'''
welcome_message_without_search = '''Nice to meet you! I am Sigmund, your friendly AI assistant! What would you like to discuss?

<div class="message-notification">I am currently functioning as a regular chatbot. In this mode I can do <a href="/about">lots of things</a>, but I cannot read the OpenSesame documentation. Therefore, if you want help with OpenSesame, <a href="/api/conversation/new/with_search">click here</a> to turn me into an OpenSesame expert.</div>'''
# The default title of a new conversation
default_conversation_title = 'New conversation'
# The number of previous messages for which tool results should be 
# retained.
keep_tool_results = 4
# Tool results larger than this are not included in the prompt used for
# generating
large_tool_result_length = 50000

# RATE LIMITS
#
# The maximum number of tokens that can be consumed per hour by the answer
# model.
max_tokens_per_hour = 500000
max_tokens_per_hour_exceeded_message = 'You have reached the hourly usage limit. Please wait and try again later!'


# LOGGING
#
# When set to True, replies will be logged. This should disabled during 
# production for privacy.
log_replies = False

# MODELS
#
# The API keys should not be shared
openai_api_key = os.environ.get('OPENAI_API_KEY', None)
anthropic_api_key = os.environ.get('ANTHROPIC_API_KEY', None)
mistral_api_key = os.environ.get('MISTRAL_API_KEY', None)
# Supported models are currently:
# - gpt-3.5
# - gpt-4o
# - gpt-4o-mini
# - claude-2.1
# - claude-3-sonnet
# - claude-3-opus
# - claude-3.5-sonnet
# - mistral-tiny
# - mistral-small
# - mistral-medium
# - dummy
# Model configuations are combinations of models that can be used together
# - The search model is used to formulate search queries and evaluate whether
#   documents are relevant. This can be a cheap model.
# - The condense model is used to summarize message history when the 
#   conversation becomes too long. This can also be a cheap model.
# - The answermodel generates the actual answer. This should be a very capable
#   model
model_config = {
    'openai': {
        'search_model': 'gpt-4o-mini',
        'condense_model': 'gpt-4o-mini',
        'public_model': 'gpt-4o-mini',
        'answer_model': 'gpt-4'
    },
    'anthropic': {
        'search_model': 'claude-3.5-sonnet',
        'condense_model': 'claude-3.5-sonnet',
        'public_model': 'claude-3-haiku',
        'answer_model': 'claude-3.5-sonnet'
    },
    'mistral': {
        'search_model': 'mistral-large-latest',
        'condense_model': 'open-mistral-nemo',
        'public_model': 'gpt-4o-mini',
        'answer_model': 'mistral-large-latest'
    },
    'dummy': {
        'search_model': 'dummy',
        'condense_model': 'dummy',
        'public_model': 'dummy',
        'answer_model': 'dummy'
    }
}
# Model-specific keyword arguments that are passed to the message generation
# functions
anthropic_kwargs = {
    'max_tokens': 1024
}
openai_kwargs = {}
mistral_kwargs = {}

# TOOLS
# 
# Tools should match the names of classes from sigmund.tools
# Search tools are executed in the first documentation-search phase
search_tools = ['search_documentation']
# Answer tools are executed during the answer phase
answer_tools_with_search = []
answer_tools_without_search = ['read_attachment', 'search_google_scholar',
                               'execute_code', 'download', 'generate_image']

# SETTINGS
#
# These are default values for settings, which are stored in the database.
settings_default = {
    # Indicates whether documentation should be searched before answering the 
    'search_first': 'true',
    # Indicates the model configuration as specified above
    'model_config': 'openai'
}

# DOCUMENTATION
#
# Jow the search-first option appears in the menu
search_first_menu_label = 'OpenSesame expert'
# Topic sources are used to feed in specific chunks of documentation that are
# relevant to a topic.
topic_sources = {
    'opensesame': 'sources/topics/opensesame.md',
    'python': 'sources/topics/inline_script.py',
    'inline_script': 'sources/topics/inline_script.py',
    'osweb': 'sources/topics/inline_javascript.js',
    'javascript': 'sources/topics/inline_javascript.js',
    'inline_javascript': 'sources/topics/inline_javascript.js',
    'datamatrix': 'sources/topics/datamatrix.py',
    'data_analysis': 'sources/topics/datamatrix.py',
    'questions_howto': 'sources/topics/questions-how-to.md',
    'mouse_response_item': 'sources/topics/mouse_response_item.md',
    'feedback_item': 'sources/topics/feedback_item.md',
    'keyboard_response_item': 'sources/topics/keyboard_response_item.md',
    'logger_item': 'sources/topics/logger_item.md',
    'loop_item': 'sources/topics/loop_item.md',
    'sampler_item': 'sources/topics/sampler_item.md',
    'sketchpad_item': 'sources/topics/sketchpad_item.md',
    'synth_item': 'sources/topics/synth_item.md',
    'sequence_item': 'sources/topics/sequence_item.md',
}
# The number of documents that is considered for each search query
search_docs_per_query = 20
# The number of documents that is kept for all search queries combined
search_docs_max = 6
# The distance metric to search through documentation. Should be 'cosine' or
# 'euclidean_distance'
search_docs_distance_metric = 'cosine'
# The embedding model used for search. This has the name of an OpenAI embedding
# model
search_embedding_model = 'text-embedding-3-large'
# The number of documents and maximum number of words that is kept for public 
# search
public_search_docs_max = 6
public_search_max_doc_length = 1000
# The distance metric used for search. The cosine metric is useful because it
# is somewhat invariant to changes in document length
search_metric = 'cosine'
# The cache folder for the library that contains documentation
db_cache = 'default'
db_cache_sources = {
    'default': '.db.cache',
    'public-with-forum': '.public-with-forum.db.cache',
    'public-without-forum': '.public-without-forum.db.cache'
}


def process_ai_message(msg):
    # This pattern looks for a colon possibly followed by any number of 
    # whitespaces # and/or HTML tags, followed by a newline and a dash, and 
    # replaces it with a colon, newline, newline, and dash
    pattern = r':\s*(<[^>]+>\s*)?\n-'
    replacement = ':\n\n-'
    return re.sub(pattern, replacement, msg)
    

def validate_user(username, password):
    """user_validation.validate() should connect to an authentication system
    that verifies the account. 
    
    Parameters
    ----------
    username : str
        Lowercase. Whitespace has been stripped.
    password : str
        Whitespace has been stripped.
        
    Returns
    -------
    str or None:
        If the user is validated, the username is returned. This is typically
        equal to the `username` parameter, but in some cases it may differ, for
        example if the user logs in with an email address instead of the actual
        username. If the user is not validated, None is returned.
    """
    try:
        import user_validation
    except ImportError:
        logger.info('no user validation script found')
        return username
    logger.info('using validation script')
    return user_validation.validate(username, password)


login_failed_message = '__User name or password incorrect. Please try again.__'


# SUBSCRIPTIONS
#
# Enable this to activate the Stripe-based subscription functionality.
subscription_required = bool(
    int(os.environ.get('SIGMUND_SUBSCRIPTION_REQUIRED', 0)))
# This is the duration of the subscription in days. This should be set to a bit
# longer than a month to provide a grace period in case of payment issues.
subscription_length = 40
#
# Stripe is a payment portal. The information below should not be shared
#
# The Price ID identifies the product. You can find it in the product catalog
# of the Stripe dashboard.
stripe_price_id = os.environ.get('STRIPE_PRICE_ID', None)
# The publishable key is the API keys section of the Developers dashboard
stripe_publishable_key = os.environ.get('STRIPE_PUBLISHABLE_KEY', None)
# The secret (API) key is the API keys section of the Developers dashboard
stripe_secret_key = os.environ.get('STRIPE_SECRET_KEY', None)
# The webhook secret is used to ensure that webhook calls actually come from
# stripe
stripe_webhook_secret = os.environ.get('STRIPE_WEBHOOK_SECRET', None)
# Additional keywords passed to stripe.checkout.Session.create to customize
# the checkout process
stripe_checkout_keywords = dict(
    allow_promotion_codes=True,
    payment_method_types=['card', 'ideal', 'bancontact', 'paypal'])


# GOOGLE LOGIN
#
google_login_enabled = True
google_client_id = os.environ.get("GOOGLE_CLIENT_ID", None)
google_client_secret = os.environ.get("GOOGLE_CLIENT_SECRET", None)
google_discovery_url = (
    "https://accounts.google.com/.well-known/openid-configuration"
)



from langchain_openai import OpenAIEmbeddings
from langchain_community.vectorstores import FAISS
from datamatrix import functional as fnc
from pathlib import Path
import logging
import openai
import time
from . import config
logger = logging.getLogger('sigmund')


class CachedEmbeddings(OpenAIEmbeddings):
    """Uses memoization to avoid calculating embeddings over and over again"""
    def embed_documents(self, texts):
        @fnc.memoize(persistent=True)
        def inner(texts):
            return super(CachedEmbeddings, self).embed_documents(texts)
        return inner(texts)


def load_library(force_reindex=False, cache_folder=config.db_cache,
                 exclude_filter=[]):
    db_cache = Path(config.db_cache_sources[cache_folder])
    src_path = Path('sources')
    embeddings_model = CachedEmbeddings(openai_api_key=config.openai_api_key,
                                        model=config.search_embedding_model)
    if not force_reindex and db_cache.exists():
        logger.info('loading library from cache')
        db = FAISS.load_local(db_cache, embeddings_model,
                              allow_dangerous_deserialization=True)
        return db.as_retriever()
    from langchain_community.document_loaders import TextLoader, \
        PyPDFLoader, JSONLoader
    logger.info('initializing library')
    data = []
    # PDF files are unstructured. They can be named through config.sources
    for src in src_path.glob('pdf/**/*.pdf'):
        if any(f in str(src) for f in exclude_filter):
            logger.info(f'skipping pdf: {src}')
            continue
        logger.info(f'indexing pdf: {src}')
        data += PyPDFLoader(str(src)).load_and_split()
    # jsonl is mainly for documentation
    for src in src_path.glob('jsonl/*.jsonl'):
        logger.info(f'indexing json: {src}')
        if any(f in str(src) for f in exclude_filter):
            logger.info(f'skipping json: {src}')
            continue            
        loader = JSONLoader(src, jq_schema='.', content_key='content',
                            json_lines=True,
                            metadata_func=_extract_metadata)
        data += loader.load()
    # To avoid running into rate limits, we throttle the ingestion of the
    # documents
    for i in range(0, len(data), config.chunk_size):
        logger.info(
            f'ingesting chunk {i}-{i + config.chunk_size}/{len(data)}')
        chunk = data[i:i + config.chunk_size]
        if not i:
            db = FAISS.from_documents(chunk, embeddings_model)
        else:
            time.sleep(config.chunk_throttle)
            db.add_documents(chunk)
    db.save_local(db_cache)
    logger.info(f'libary initialized')
    return db.as_retriever()


def _extract_metadata(record, metadata):
    metadata['url'] = record.get('url', record.get('source', None))
    metadata['title'] = record['title']
    return metadata



import json
import logging
import time
import base64
from datetime import datetime, timedelta
from .. import config, attachments
from .models import db, User, Conversation, Attachment, Activity, \
    Subscription, Setting, Message
from .encryption import EncryptionManager
from sqlalchemy import func
from sqlalchemy.orm.exc import NoResultFound

logger = logging.getLogger('sigmund')
ATTACHMENT_PREFIX = '::plain-text-attachment::'


class DatabaseManager:
    
    def __init__(self, sigmund, username: str,
                 encryption_key: [str, bytes]=None):
        self._sigmund = sigmund
        self.username = username
        self.encryption_manager = EncryptionManager(encryption_key)
        self.ensure_user_exists()
        logger.info(
            f'initializing database for user {self.username} ({self.user_id})')

    def ensure_user_exists(self):
        try:
            user = User.query.filter_by(username=self.username).one()
            self.user_id = user.user_id
        except NoResultFound:
            logger.info(f'creating new user: {self.username}')
            user = User(username=self.username)
            db.session.add(user)
            db.session.commit()
            self.user_id = user.user_id
            self.new_conversation()
    
    def get_message(self, message_id: int) -> dict:
        # If it's not an int, then it's an old-style mesage
        if not isinstance(message_id, int):
            return message_id
        try:
            # First, check if the message exists in the new Message table
            message = Message.query.filter_by(message_id=message_id).first()
            if message:
                # If it exists, decrypt and return the message
                decrypted_data = self.encryption_manager.decrypt_data(message.data)
                return json.loads(decrypted_data)
            else:
                # If it doesn't exist, assume it's an old-style message
                # embedded in conversation data, and return as is
                return message_id
        except Exception as e:
            logger.error(f"Error retrieving message {message_id}: {e}")
            return {}
    
    def add_message(self, conversation_id: int, message_data: dict) -> int:
        json_data = json.dumps(message_data)
        encrypted_data = self.encryption_manager.encrypt_data(
            json_data.encode('utf-8'))
        message = Message(conversation_id=conversation_id, data=encrypted_data)
        db.session.add(message)
        db.session.commit()
        return message.message_id
    
    def get_active_conversation(self) -> dict:
        try:
            user = User.query.filter_by(user_id=self.user_id).one()
            conversation = Conversation.query.filter_by(
                conversation_id=user.active_conversation_id).one()
            decrypted_data = self.encryption_manager.decrypt_data(conversation.data)
            conversation_data = json.loads(decrypted_data)
    
            # Fetch and decrypt messages
            message_ids = conversation_data.get('message_history', [])
            message_history = [self.get_message(msg_id) for msg_id in message_ids]
            conversation_data['message_history'] = message_history
    
            return conversation_data
        except NoResultFound:
            logger.warning(f"No active conversation found for user {self.user_id}")
            return {}    

    def set_active_conversation(self, conversation_id: int) -> bool:
        try:
            # We first get the active conversation
            conversation = Conversation.query.filter_by(
                conversation_id=conversation_id, user_id=self.user_id).one()
            # And then update the current time for that conversation so that
            # it ends on top
            decrypted_data = self.encryption_manager.decrypt_data(
                conversation.data)
            conversation_data = json.loads(decrypted_data)
            conversation_data['last_updated'] = time.time()
            json_data = json.dumps(conversation_data)
            encrypted_data = self.encryption_manager.encrypt_data(
                json_data.encode('utf-8'))
            conversation.data = encrypted_data
            # Finally we change the active conversation for the user
            user = User.query.filter_by(user_id=self.user_id).one()
            user.active_conversation_id = conversation.conversation_id
            db.session.commit()
            return True
        except NoResultFound:
            logger.warning(f"Conversation {conversation_id} not found or does "
                           f"not belong to user {self.user_id}")
            return False
    
    def update_active_conversation(self, conversation_data: dict) -> bool:
        try:
            user = User.query.filter_by(user_id=self.user_id).one()
            conversation = Conversation.query.filter_by(
                conversation_id=user.active_conversation_id).one()
    
            # Extract messages and store them separately
            message_history = conversation_data.pop('message_history', [])
            message_ids = []
            for message in message_history:
                message_id = self.add_message(conversation.conversation_id, message)
                message_ids.append(message_id)
    
            conversation_data['message_history'] = message_ids
            conversation_data['last_updated'] = time.time()
    
            json_data = json.dumps(conversation_data)
            encrypted_data = self.encryption_manager.encrypt_data(json_data.encode('utf-8'))
            conversation.data = encrypted_data
            db.session.commit()
            return True
        except NoResultFound:
            logger.warning(f"No active conversation to update for user {self.user_id}")
            return False    

    def list_conversations(self) -> dict:
        
        conversations = {}
        user = User.query.filter_by(user_id=self.user_id).one()
        for conversation in \
                Conversation.query.filter_by(user_id=self.user_id).all():
            try:
                data = json.loads(self.encryption_manager.decrypt_data(
                    conversation.data))
                # Don't list empty conversations except for the active one
                if len(data.get('message_history', [])) < 2 and \
                        user.active_conversation_id != conversation.conversation_id:
                    continue
                conversations[conversation.conversation_id] = (
                     data.get('title', 'Untitled conversation'),
                     data.get('last_updated', time.time()))
            except Exception as e:
                logger.error(f"Error decrypting conversation data: {e}")
        return conversations

    def new_conversation(self) -> bool:
        try:
            conversation_data = {
                "title": "New conversation",
                "condensed_text": None,
                "message_history": [],
                "message_metadata": [],
                "last_updated": time.time()
            }
            json_data = json.dumps(conversation_data)
            encrypted_data = self.encryption_manager.encrypt_data(
                json_data.encode('utf-8'))
            conversation = Conversation(user_id=self.user_id,
                                        data=encrypted_data)
            db.session.add(conversation)
            db.session.commit()
            self.set_active_conversation(conversation.conversation_id)
            return True
        except Exception as e:
            logger.error(f"Error creating new conversation: {e}")
            return False
    
    def delete_conversation(self, conversation_id: int) -> bool:
        try:
            user = User.query.filter_by(user_id=self.user_id).one()
            conversation = Conversation.query.filter_by(
                conversation_id=conversation_id, user_id=self.user_id).one()
            if user.active_conversation_id == conversation.conversation_id:
                logger.warning(
                    f"Cannot delete active conversation {conversation_id} for "
                    f"user {self.user_id}")
                return False

            # Delete associated messages
            Message.query.filter_by(conversation_id=conversation_id).delete()

            # Delete the conversation
            db.session.delete(conversation)
            db.session.commit()
            return True
        except NoResultFound:
            logger.warning(f"Conversation {conversation_id} not found or does "
                           f"not belong to user {self.user_id}")
            return False    

    def list_attachments(self) -> dict:
        attachments = Attachment.query.filter_by(user_id=self.user_id).all()
        attachment_dict = {}
        for attachment in attachments:
            try:
                decrypted_data = self.encryption_manager.decrypt_data(
                    attachment.data)
                attachment_data = json.loads(decrypted_data)
                attachment_dict[attachment.attachment_id] = {
                    'filename' : attachment_data.get(
                        'filename', 'No filename'),
                    'description' : attachment_data.get(
                        'description', 'No description')
                }
            except Exception as e:
                logger.error(f"Error decrypting attachment data for "
                             f"attachment_id {attachment.attachment_id}: {e}")
        return attachment_dict

    def delete_attachment(self, attachment_id: int) -> bool:
        try:
            attachment = Attachment.query.filter_by(
                attachment_id=attachment_id, user_id=self.user_id).one()
            db.session.delete(attachment)
            db.session.commit()
            return True
        except NoResultFound:
            logger.warning(f"Attachment {attachment_id} not found or does not "
                           f"belong to user {self.user_id}")
            return False

    def add_attachment(self, attachment_data: dict) -> int:
        """Adds an attachment and returns the attachment_id or -1 if an error
        occurred. attachment_data should be a dict with filename, content,
        and description keys, where content is a base64-encoded str.
        """
        try:
            attachment_data['plaintext'] = True
            json_data = json.dumps(attachment_data)
            encrypted_data = self.encryption_manager.encrypt_data(
                json_data.encode('utf-8'))
            attachment = Attachment(user_id=self.user_id, data=encrypted_data)
            db.session.add(attachment)
            db.session.commit()
            return attachment.attachment_id
        except Exception as e:
            breakpoint()
            logger.error(f"Error adding attachment: {e}")
            return -1

    def get_attachment(self, attachment_id: int) -> dict:
        try:
            attachment_record = Attachment.query.filter_by(
                attachment_id=attachment_id, user_id=self.user_id).one()
            attachment = json.loads(
                self.encryption_manager.decrypt_data(attachment_record.data))
            # Old attachments are stored as binary blobs and need to be
            # explicitly converted to text. These can be recognized by the
            # fact that they do not start with the attachment prefix
            if not attachment.get('plaintext', False):
                logger.info('attachment stored as binary, decoding')
                attachment['content'] = attachments.file_to_text(
                    attachment['filename'],
                    base64.b64decode(attachment['content']),
                    self._sigmund.condense_model)
            return attachment
        except NoResultFound:
            logger.warning(f"Attachment {attachment_id} not found or does not "
                           f"belong to user {self.user_id}")
            return {}
        except Exception as e:
            logger.error(f"Error decrypting attachment data for attachment_id "
                         f"{attachment_id}: {e}")
            return {}

    def add_activity(self, tokens_consumed: int):
        """Adds the number of tokens consumed using the current time."""
        new_activity = Activity(user_id=self.user_id, time=datetime.utcnow(),
                                tokens_consumed=tokens_consumed)
        db.session.add(new_activity)
        db.session.commit()

    def get_activity(self, after_time=None) -> int:
        """Retrieves the total number of tokens consumed since a particular 
        time, using the past hour as a default if no after_time is provided.
        """
        if after_time is None:
            after_time = datetime.utcnow() - timedelta(hours=1)

        total_tokens = db.session.query(func.sum(Activity.tokens_consumed)) \
            .filter(Activity.user_id == self.user_id) \
            .filter(Activity.time >= after_time) \
            .scalar()
        return total_tokens if total_tokens is not None else 0

    def update_subscription(self, stripe_customer_id: str,
                            stripe_subscription_id: str, from_date=None,
                            to_date=None):
        """Updates a subscription. By default the subscriptions starts 
        immediately and ends exactly one month from the current time. A user
        has only one subscription, which means that a subscription should be
        added if it doesn't exist, and updated if it already exists.
        """
        now = datetime.utcnow()
        from_date = from_date or now
        to_date = to_date or now + timedelta(days=config.subscription_length)
        subscription = Subscription.query.filter_by(
            user_id=self.user_id).first()
        if subscription:
            subscription.from_date = from_date
            subscription.to_date = to_date
        else:
            subscription = Subscription(
                user_id=self.user_id, from_date=from_date, to_date=to_date,
                stripe_customer_id=stripe_customer_id,
                stripe_subscription_id=stripe_subscription_id)
            db.session.add(subscription)
        db.session.commit()

    def check_subscription(self) -> bool:
        """Returns a bool indicating whether the user is currently subscribed.
        """
        now = datetime.utcnow()
        subscription = Subscription.query.filter_by(
            user_id=self.user_id).first()
        print(now)
        print(subscription.from_date)
        print(subscription.to_date)
        return subscription and \
            subscription.from_date <= now < subscription.to_date

    def cancel_subscription(self):
        """Sets the to_date of the subscription to the current time if the user
        is currently subscribed.
        """
        if not self.check_subscription():
            return
        subscription = Subscription.query.filter_by(
            user_id=self.user_id).first()
        subscription.to_date = datetime.utcnow()
        db.session.commit()

    def get_stripe_customer_id(self) -> str:
        subscription_record = Subscription.query.filter(
            Subscription.user_id == self.user_id
        ).order_by(Subscription.from_date.desc()).first()
        if subscription_record:
            return subscription_record.stripe_customer_id
        return None

    @staticmethod
    def from_stripe_customer_id(stripe_customer_id: str) -> str:
        subscription_record = Subscription.query.filter(
            Subscription.stripe_customer_id == stripe_customer_id
        ).order_by(Subscription.from_date.desc()).first()
        if not subscription_record:
            return None
        user_record = User.query.filter(
            User.user_id == subscription_record.user_id
        ).first()
        if not user_record:
            return None
        return DatabaseManager(None, username=user_record.username)

    def get_setting(self, key: str) -> str:
        """Retrieve a setting value for the current user, which is available
        as self.user_id. If the setting does not exist, return the default
        value as specified in the config or None if no default has been 
        specified.
        """
        setting = db.session.query(Setting).filter_by(
            user_id=self.user_id, key=key).first()
        return setting.value if setting \
            else config.settings_default.get(key, None)

    def set_setting(self, key: str, value: str):
        """Set a setting to specified value for the current user. If the
        setting already exists, overwrite it.
        """
        setting = Setting.query.filter_by(
            user_id=self.user_id, key=key).first()
        if setting:
            setting.value = value
        else:
            new_setting = Setting(user_id=self.user_id, key=key,
                                  value=value)
            db.session.add(new_setting)
        db.session.commit()

    def list_settings(self) -> dict:
        """Returns a dict of all settings for the current user."""
        settings = {}
        setting_objs = db.session.query(Setting).filter_by(
            user_id=self.user_id).all()
        for setting in setting_objs:
            settings[setting.key] = setting.value
        return settings



import logging
import jinja2
import json
from types import GeneratorType
from typing import Tuple, Optional
from . import config, library
from .documentation import Documentation, FAISSDocumentationSource
from .messages import Messages
from .model import model
from .database.manager import DatabaseManager
from . import prompt
from . import tools
logger = logging.getLogger('sigmund')


class Sigmund:
    """The main chatbot class.
    
    Parameters
    ----------
    user_id: A user name
    persistent: Indicates whether the current session should be persistent in
        the sense of using the database.
    search_first: Indicates whether the answer phase should be preceded by a
        documentation search phase. If None, the config default will be used.
    model_config: Indicates the model configuration. If none, the config
        default will be used.
    search_tools: A list of tools to be used by the documentation search (if
        enabled). Values are class names from sigmund.tools. If None, the
        config default will be used.
    answer_tools: A list of tools to be used during the answer phase (if
        enabled). Values are class names from sigmund.tools. If None, the
        config default will be used.
    """
    def __init__(self, user_id: str, persistent: bool = False,
                 encryption_key: Optional[str] = None,
                 search_first: Optional[bool] = None,
                 model_config: Optional[str] = None,
                 search_tools: Optional[list] = None,
                 answer_tools: Optional[list] = None):
        self.user_id = user_id
        self.database = DatabaseManager(self, user_id, encryption_key)
        # Search first is stored as a str but should be a bool here
        self.search_first = (
            self.database.get_setting('search_first') == 'true'
            if search_first is None else search_first)
        self.model_config = config.model_config[
            self.database.get_setting('model_config')
            if model_config is None else model_config
        ]
        self.documentation = Documentation(
            self, sources=[FAISSDocumentationSource(self)])
        self.messages = Messages(self, persistent)
        if search_tools is None:
            search_tools = config.search_tools
        if answer_tools is None:
            if self.search_first:
                answer_tools = config.answer_tools_with_search
            else:
                answer_tools = config.answer_tools_without_search
        # Tools are class names from the tools module, which need to be
        # instantiated with sigmund (self) as first argument
        self.search_tools = [getattr(tools, t)(self) for t in search_tools]
        self.answer_tools = [getattr(tools, t)(self) for t in answer_tools]
        # If there are search tools, the first one should always be used
        if search_tools:
            search_tool_choice = search_tools[0]
        else:
            search_tool_choice = None
        # If there are answer tools, the mode can choose freely
        if answer_tools:
            answer_tool_choice = 'auto'
        else:
            answer_tool_choice = None
        self.search_model = model(self, self.model_config['search_model'],
                                  tools=self.search_tools, 
                                  tool_choice=search_tool_choice)
        self.answer_model = model(self, self.model_config['answer_model'],
                                  tools=self.answer_tools,
                                  tool_choice=answer_tool_choice)
        self.condense_model = model(self, self.model_config['condense_model'])
        self.public_model = model(self, self.model_config['public_model'])
    
    def send_user_message(self, message: str,
                          message_id: str=None) -> GeneratorType:
        """The main function that takes a user message and returns one or 
        replies. This is a generator function where each yield gives a tuple.
        
        This tuple can be (dict, dict) in which case the first dict contains an 
        action and a message key. This communicates to the client that an 
        action should be taking, such that the loading indicator should change.
        This tuple can also be (str, dict) in which case the str contains an
        AI message and the dict contains metadata for the message.
        """
        if self._rate_limit_exceeded():
            yield config.max_tokens_per_hour_exceeded_message, \
                    self.messages.metadata()
            return
        self.messages.append('user', message, message_id)
        if self.search_first:
            for reply, metadata in self._search(message):
                yield reply, metadata
        for reply, metadata in self._answer():
            yield reply, metadata
            
    def _rate_limit_exceeded(self):
        tokens_consumed_past_hour = self.database.get_activity()
        logger.info(
            f'tokens consumed in past hour: {tokens_consumed_past_hour}')
        return tokens_consumed_past_hour > config.max_tokens_per_hour
    
    def _search(self, message: str) -> GeneratorType:
        """Implements the documentation search phase."""
        yield {'action': 'set_loading_indicator',
               'message': f'{config.ai_name} is searching '}, {}
        logger.info('[search state] entering')
        self.documentation.clear()
        # First seach based on the user question
        logger.info('[search state] searching based on user message')
        self.documentation.search([message])
        # Then search based on the search-model queries derived from the user
        # question
        reply = self.search_model.predict(self.messages.prompt(
            system_prompt=prompt.SYSTEM_PROMPT_SEARCH,
            skip_large_tool_results=True))
        if config.log_replies:
            logger.info(f'[search state] reply: {reply}')
        if callable(reply):
            reply()
        else:
            logger.warning(f'[search state] did not call search tool')
        self.documentation.strip_irrelevant(message)
        logger.info(
            f'[search state] {len(self.documentation._documents)} documents, {len(self.documentation)} characters')
    
    def _answer(self, state: str = 'answer') -> GeneratorType:
        """Implements the answer phase."""
        yield {'action': 'set_loading_indicator',
               'message': f'{config.ai_name} is thinking and typing '}, {}        
        logger.info(f'[{state} state] entering')
        # We first collect a regular reply to the user message. While doing so
        # we also keep track of the number of tokens consumed.
        tokens_consumed_before = self.answer_model.total_tokens_consumed
        reply = self.answer_model.predict(self.messages.prompt(
            skip_large_tool_results=True))
        tokens_consumed = self.answer_model.total_tokens_consumed \
            - tokens_consumed_before
        logger.info(f'tokens consumed: {tokens_consumed}')
        self.database.add_activity(tokens_consumed)        
        if config.log_replies:
            logger.info(f'[{state} state] reply: {reply}')
        # If the reply is a callable, then it's a tool that we need to run
        if callable(reply):
            tool_message, tool_result, needs_feedback = reply()
            if needs_feedback:
                logger.info(f'[{state} state] tools need feedback')
                if not self.answer_model.supports_tool_feedback:
                    logger.info(
                        f'[{state} state] model does not support feedback')
                    needs_feedback = False
            metadata = self.messages.append('assistant', tool_message)
            yield tool_message, metadata
            # If the tool has a result, yield and remember it
            if tool_result:
                metadata = self.messages.append('tool',
                                                json.dumps(tool_result))
                if tool_result['content']:
                    yield tool_result['content'], metadata
        # Otherwise the reply is a regular AI message
        else:
            metadata = self.messages.append('assistant', reply)
            yield reply, metadata
            # If the reply contains a NOT_DONE_YET marker, this is a way for the AI
            # to indicate that it wants to perform additional actions. This makes 
            # it easier to perform tasks consisting of multiple responses and 
            # actions. The marker is stripped from the reply so that it's hidden
            # from the user. We also check for a number of common linguistic 
            # indicators that the AI isn't done yet, such "I will now". This is
            # necessary because the explicit marker isn't reliably sent.
            if self.answer_model.supports_not_done_yet and \
                    prompt.NOT_DONE_YET_MARKER in reply:
                reply = reply.replace(prompt.NOT_DONE_YET_MARKER, '')
                logger.info(f'[{state} state] not-done-yet marker received')
                needs_feedback = True
            else:
                needs_feedback = False
        # If feedback is required, either because the tools require it or 
        # because the AI sent a NOT_DONE_YET marker, go for another round.
        if needs_feedback and not self._rate_limit_exceeded():
            for reply, metadata in self._answer(state='feedback'):
                yield reply, metadata



from . import BaseTool
from .. import config
import logging
from pathlib import Path
from langchain_core.documents import Document
logger = logging.getLogger('sigmund')


class search_documentation(BaseTool):
    """Search the documentation based on topics and search queries"""
    
    arguments = {
        "primary_topic": {
            "type": "string",
            "description": "The primary topic of the question"
        },
        "secondary_topic": {
            "type": "string",
            "description": "The secondary topic of the question"
        },
        "queries": {
            "type": "array",
            "items": {
                "type": "string"
            },
            "description": "A list of queries to search the documentation",
        }
    }
    required_arguments = ['primary_topic', 'queries']
    
    @property
    def tool_spec(self):
        topics = list(config.topic_sources.keys())
        arguments = self.arguments.copy()
        arguments['primary_topic']['enum'] = topics
        arguments['secondary_topic']['enum'] = topics
        return {
            "name": self.__class__.__name__,
            "description": self.__doc__,
            "parameters": {
                "type": "object",
                "properties": arguments,
                "required": self.required_arguments,
            }
        }    
    
    def __call__(self, primary_topic, queries, secondary_topic=None):
        topics = [primary_topic]
        if secondary_topic:
            topics.append(secondary_topic)
        for topic in topics:
            if topic not in config.topic_sources:
                logger.warning(f'unknown topic: {topic}')
                continue
            logger.info(f'appending doc for topic: {topic}')
            doc = Document(
                page_content=Path(config.topic_sources[topic]).read_text())
            doc.metadata['important'] = True
            self._sigmund.documentation.append(doc)
        self._sigmund.documentation.search(queries)
        return 'Searching documentation ...', '', False



from .. import config, utils
from . import BaseModel
from ._openai_model import OpenAIModel
import logging
from langchain.schema import SystemMessage, AIMessage, HumanMessage, \
    FunctionMessage
logger = logging.getLogger('sigmund')


class MistralModel(OpenAIModel):
    
    supports_not_done_yet = False

    def __init__(self, sigmund, model, **kwargs):
        from mistralai.async_client import MistralAsyncClient
        from mistralai.client import MistralClient
        BaseModel.__init__(self, sigmund, **kwargs)
        self._model = model
        # Mistral doesn't allow a tool to be specified by name. So if this
        # happens, we instead use the 'any' option, which forces use of the
        # best fitting tool, which in the case of a single tool boils down to
        # the same thing as forcing the tool by name.
        if self._tool_choice not in [None, 'none', 'auto', 'any']:
            self._tool_choice = 'any'
        self._client = MistralClient(api_key=config.mistral_api_key)
        self._async_client = MistralAsyncClient(api_key=config.mistral_api_key)
        
    def predict(self, messages):
        if isinstance(messages, str):
            messages = [self.convert_message(messages)]
        else:
            messages = utils.prepare_messages(messages, allow_ai_first=False,
                                              allow_ai_last=False,
                                              merge_consecutive=True)
            messages = [self.convert_message(message) for message in messages]
            messages = self._prepare_tool_messages(messages)
        # Mistral requires an assistant message after a tool message
        while True:
            for i, message in enumerate(messages[:-1]):
                next_message = messages[i + 1]
                if message['role'] == 'tool' and \
                        next_message['role'] == 'user':
                    break
            else:
                break
            logger.info('adding assistant message between tool and user')
            messages.insert(i + 1, {'role': 'assistant',
                                'content': 'Tool was executed.'})
        return BaseModel.predict(self, messages)
        
    def _tool_call_id(self, nr):
        # Must be a-z, A-Z, 0-9, with a length of 9
        return f'{nr:09d}'
        
    def _mistral_invoke(self, fnc, messages):
        # Mistral tends to get stuck in a loop where the same tool is called
        # over and over again. To fix this, we temporarily disallow tools when
        # the last message was a tool.	
        if messages[-1]['role'] == 'tool':
            kwargs = {}
        else:
            kwargs = self._tool_args()        
        kwargs.update(config.mistral_kwargs)
        if self.json_mode:
            kwargs['response_format'] = {"type": "json_object"}
        return fnc(model=self._model, messages=messages, **kwargs)
    
    def invoke(self, messages):
        return self._mistral_invoke(self._client.chat, messages)
        
    def async_invoke(self, messages):
        return self._mistral_invoke(self._async_client.chat, messages)



from . import BaseTool
import logging
import requests
from .. import config
logger = logging.getLogger('sigmund')


class generate_image(BaseTool):
    """Generate an image based on a text description."""
    
    arguments = {
        'prompt': {
            'type': 'string',
            'description': 'The image description',
        },
        'quality': {
            'type': 'string',
            'description': 'The quality of the image. Use standard unless there is a good reason to use HD.',
            'enum': ['standard', 'hd']
        },
        'size': {
            'type': 'string',
            'description': 'The size of the image in pixels',
            'enum': ['1024x1024', '1792x1024', '1024x1792']
        },
        'style': {
            'type': 'string',
            'description': 'Vivid causes the model to lean towards generating hyper-real and dramatic images. Natural causes the model to produce more natural, less hyper-real looking images.',
            'enum': ['natural', 'vivid']
        }
    }
    required_arguments = ['prompt']
    
    def __call__(self, prompt, quality='standard', size='1024x1024',
                 style='natural'):
        from openai import OpenAI
        logger.info(f'generating image: {prompt}, quality={quality}, '
                    f'size={size}, style={style}')
        client = OpenAI(api_key=config.openai_api_key)
        response = client.images.generate(model='dall-e-3', prompt=prompt,
                                          size=size, quality=quality,
                                          style=style,
                                          response_format='b64_json', n=1)
        data = response.data[0].b64_json
        html = f'Here\'s an image for "{prompt}"!\n<div class="image-generation"><img src="data:image/png;base64,{data}"></div>'
        return f'Generating image ...', html, False



import jinja2

# The system prompt during documentation search consists of the prompt below
SYSTEM_PROMPT_SEARCH = '''Do not answer the user's question. Instead, use the search_documentation function tool to search for relevant documentation.'''

# The system prompt used during question answering is composed of the fragments
# below
SYSTEM_PROMPT_IDENTITY_WITH_SEARCH = '''You are Sigmund, a brilliant AI assistant for users of OpenSesame, a program for building psychology and neuroscience experiments. Your knowledge of OpenSesame is outdated. Therefore, strictly follow the documentation that is provided in the context between <documentation> tags. You sometimes use emojis.'''
SYSTEM_PROMPT_IDENTITY_WITHOUT_SEARCH = '''You are Sigmund, a brilliant AI assistant. You sometimes use emojis.'''
# Sent by AI to indicate that message requires for replies or actions
NOT_DONE_YET_MARKER = '<NOT_DONE_YET>'
SYSTEM_PROMPT_NOT_DONE_YET = f'''When you intend to perform an action ("please wait", "I will now"), such as searching or code execution, end your reply with {NOT_DONE_YET_MARKER}.'''
SYSTEM_PROMPT_ATTACHMENTS = '''# Attachments

You have access to the following attached files:

{{ description }}
'''
SYSTEM_PROMPT_CONDENSED = '''Below is a summary of the start of the conversation:

<summary>
{{ summary }}
</summary>'''


CONDENSE_HISTORY = '''Summarize the following conversation:

{{ history }}
'''

JUDGE_RELEVANCE = '''Is the document relevant for answering the question?

# Question

{{ question }}
"""

# Document

<document>
{{ documentation }}
</document>

# Reply format

Reply with a JSON string to indicate whether the document is relevant ({"relevant": true}) or not ({"relevant": false}). Do not include any additional text.'''

TITLE_PROMPT = '''Provide a brief title that desribes the topic of the conversation below. Reply only with the title, do not add any additional text.'''
DESCRIBE_PROMPT = '''Provide a brief description of the following text:
        
Filename: {{ name }}
        
<TEXT>
{{ text_representation }}
</TEXT>'''
SUMMARIZE_PROMPT = '''Provide a long comprehensive summary of the following text:

<TEXT>
{{ text_representation }}
</TEXT>'''
PUBLIC_SEARCH_PROMPT = '''For each of the documentation sections below, provide a summary in a single bullet point using this format: "- [title](url): description (max 2 sentences)". Only reply with bullet points.
    
{{ documentation }}
'''


def render(tmpl, **kwargs):
    return jinja2.Template(tmpl).render(**kwargs)



from . import BaseTool
import logging
import requests
logger = logging.getLogger('sigmund')


class execute_code(BaseTool):
    """Execute Python and R code. Do not use this tool unless code should actually be executed. When executing Python, you can only use modules from the Python standard library."""
    
    arguments = {
        'language': {
            'type': 'string',
            'description': 'The programming language to use',
            'enum': ['r', 'python']
        },
        'code': {
            'type': 'string',
            'description': 'The code to execute. Use print() to print to the standard output.'
        }
    }
    required_arguments = ['language', 'code']
    
    def __call__(self, language, code):
        logger.info(f'executing {language} code: {code}')
        url = "https://emkc.org/api/v2/piston/execute"
        language_aliases = {'python': 'python',
                            'r': 'r'}
        language = language_aliases[language.lower()]
        language_versions = {'python': '3.10', 'r': '4.1.1'}
        language_files = {'python': 'main.py', 'r': 'main.R'}
        data = {
            "language": language,
            "version": language_versions[language],
            "files": [{"name": language_files[language], "content": code}],
            "stdin": "",
            "args": [],
            "compile_timeout": 10000,
            "run_timeout": 3000,
            "compile_memory_limit": -1,
            "run_memory_limit": -1
        }
        response = requests.post(url, json=data)
        if response.status_code == 200:
            response_data = response.json()
            result = response_data.get("run", {}).get("output", "").strip()
            logger.info(f'result: {result}')
            result = f'''I executed the following code:

```{language}
{code}
```

And received the following output:

```
{result}
```
'''
            return 'Executing code ...', result, True
        logger.error(f"Error: {response.status_code} with message: {response.content}")
        return 'Failed to execute code', '', True



import html
import logging
import textwrap
import time
from flask import render_template, render_template_string
import markdown
from markdown.extensions.fenced_code import FencedCodeExtension
from markdown.extensions.codehilite import CodeHiliteExtension
from markdown.extensions.toc import TocExtension
from markdown.extensions.attr_list import AttrListExtension
from langchain.schema import HumanMessage
from . import config
from . import __version__
logger = logging.getLogger('sigmund')


def md(text):
    if '<REPORTED>' in text or '<FINISHED>' in text:
        return text
    return markdown.markdown(text,
                             extensions=[FencedCodeExtension(),
                                         CodeHiliteExtension(),
                                         TocExtension(),
                                         AttrListExtension()])


def clean(text, escape_html=True, render=True):
    if render:
        text = render_template_string(text)
    if escape_html:
        text = html.escape(text)
    return text
    

def render(path, **kwargs):
    return render_template(
        path,
        ai_name=config.ai_name,
        page_title=config.page_title,
        server_url=config.server_url,
        max_message_length=config.max_message_length,
        version=__version__,
        **kwargs)


def deindent_code_blocks(text):
    in_block = False
    lines = []
    for line_nr, line in enumerate(text.splitlines()):
        if in_block:
            block.append(line)
            # Ending line
            if line.lstrip().startswith('```'):
                in_block = False
                block = textwrap.dedent('\n'.join(block))
                lines.append(block)
            continue
        # Starting line
        if line.lstrip().startswith('```'):
            in_block = True
            block = [line]
        else:
            lines.append(line)
    return '\n'.join(lines)
    
    
def prepare_messages(messages, allow_ai_first=True, allow_ai_last=True,
                     merge_consecutive=False, merge_separator='\n'):
    """Takes a list of Message objects formats them to meet the requirements
    of specific models.
    
    Parameters
    ----------
    messages : list
        A list of messages, where each message object has a type property that
        is either 'system', 'ai', or ;'human'
    allow_ai_first : bool, optional
        Indicates whether the first message (after the system message) can be
        an AI message. If not, then it is removed.
    allow_ai_last : bool, optional
        Indicates whether the last message (after the system message) can be
        an AI message. If not, then a human message is appended after it.
    merge_consecutive : bool, optional
        Indicates whether multiple messages of the same class should be merged
        into a single message.
    merge_separator : str, optional
        If messages are merged, the separator that is used to join their 
        content.
        
    Returns
    -------
    list
    """
    if not isinstance(messages, list):
        return messages
    if not messages:
        return []
    # Check whether first message after the system message is an AI message, 
    # and remove this if not allowed
    if not allow_ai_first and messages[1].type == 'ai':
        logger.info('removing first assistant mesage')
        messages.pop(1)        
    # Merge consecutive messages of the same class totether
    if len(messages) > 1 and merge_consecutive:
        merged_messages = []
        current_message = messages[0]
        for next_message in messages[1:]:
            if current_message.type == next_message.type:
                logger.info('merging consecutive messages')
                current_message.content += merge_separator + next_message.content
            else:
                merged_messages.append(current_message)
                current_message = next_message
        merged_messages.append(current_message)
        messages = merged_messages
    # Make sure the last message is human if this is required
    if not allow_ai_last and messages[-1].type == 'ai':
        logger.info('adding continue message')
        messages.append(HumanMessage(content='Please continue!'))
    return messages


def current_datetime():
    return time.strftime('%a %d %b %Y %H:%M')



from . import BaseTool
import logging
import json
import base64
from .. import utils
from ..attachments import file_to_text
logger = logging.getLogger('sigmund')


class read_attachment(BaseTool):
    """Read an attached file. The full contents of the attachment is returned as plain text."""
    arguments = {
        "filename": {
            "type": "string",
            "description": "The attachment file to read",
        }
    }
    required_arguments = ["filename"]
    
    @property
    def tool_spec(self):
        arguments = self.arguments.copy()
        arguments['filename']['enum'] = [
            attachment['filename'] for attachment in
            self._sigmund.database.list_attachments().values()
        ]
        return {
            "name": self.__class__.__name__,
            "description": self.__doc__,
            "parameters": {
                "type": "object",
                "properties": arguments,
                "required": self.required_arguments,
            }
        }
    
    def __call__(self, filename):
        texts = []
        for attachment_id, attachment in \
                self._sigmund.database.list_attachments().items():
            if filename != attachment['filename']:
                continue
            attachment = self._sigmund.database.get_attachment(attachment_id)
            content = attachment['content']
            text = f'File name: {attachment["filename"]}\n\nFile content:\n{content}'
            result = f'''One moment please ...
        
<div class="attachment-content">
{text}
</div>'''            
            return 'I am going to read the attached file now.', result, True
        return 'Something went wrong while trying to read the attachment', \
            '', False
